"""Module implementing representational similarity analysis (RSA) at the sensor level.

Kriegeskorte, N., Mur, M., & Bandettini, P. A. (2008). Representational similarity
analysis - connecting the branches of systems neuroscience. Frontiers in Systems
Neuroscience, 2(November), 4. https://doi.org/10.3389/neuro.06.004.2008

Authors
-------
Marijn van Vliet <w.m.vanvliet@gmail.com>
"""

import numpy as np
from scipy.spatial import distance
import mne
from mne.utils import logger
from mne.cov import compute_whitener

from .rdm import _n_items_from_rdm, rdm_array
from .searchlight import searchlight
from .rsa import rsa_array


def rsa_evokeds(
    evokeds,
    rdm_model,
    noise_cov=None,
    spatial_radius=None,
    temporal_radius=None,
    evoked_rdm_metric="correlation",
    evoked_rdm_params=dict(),
    rsa_metric="spearman",
    ignore_nan=False,
    y=None,
    n_folds=1,
    picks=None,
    tmin=None,
    tmax=None,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA in a searchlight pattern on evokeds.

    The output is an Evoked object where the "signal" at each sensor is the RSA,
    computed using all surrounding sensors.

    Parameters
    ----------
    evokeds : list of :class:`mne.Evoked`
        The evoked brain activity for each item. If you have more than one Evoked object
        per item (i.e. repetitions), you can use the ``y`` parameter to assign evokeds
        to items.
    rdm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model RDM, see :func:`compute_rdm`. For efficiency, you can give it in
        condensed form, meaning only the upper triangle of the matrix as a vector. See
        :func:`scipy.spatial.distance.squareform`. To perform RSA against multiple
        models at the same time, supply a list of model RDMs.

        Use :func:`compute_rdm` to compute RDMs.
    noise_cov : :class:`mne.Covariance` | None
        When specified, the data will by normalized using the noise covariance. This is
        recommended in all cases, but a hard requirement when the data contains sensors
        of different types. Defaults to None.
    spatial_radius : float | None
        The spatial radius of the searchlight patch in meters. All sensors within this
        radius will belong to the searchlight patch. Set to None to only perform the
        searchlight over time, flattening across sensors. Defaults to None.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to only
        perform the searchlight over sensors, flattening across time. Defaults to None.
    evoked_rdm_metric : str
        The metric to use to compute the RDM for the evokeds. This can be any metric
        supported by the scipy.distance.pdist function. See also the
        ``evoked_rdm_params`` parameter to specify and additional parameter for the
        distance function. Defaults to 'correlation'.
    evoked_rdm_params : dict
        Extra arguments for the distance metric used to compute the RDMs. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing the
        distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    y : ndarray of int, shape (n_items,) | None
        For each Evoked, a number indicating the item to which it belongs. When
        ``None``, each Evoked is assumed to belong to a different item. Defaults to
        ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance metric.
        Folds are created based on the ``y`` parameter. Specify ``None`` to use the
        maximum number of folds possible, given the data. Alternatively, you can pass a
        Scikit-Learn cross validator object (e.g. ``sklearn.model_selection.KFold``) to
        assert fine-grained control over how folds are created.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted as channel
        indices. In lists, channel *type* strings (e.g., ``['meg', 'eeg']``) will pick
        channels of those types, channel *name* strings (e.g., ``['MEG0111',
        'MEG2623']`` will pick the given channels. Can also be the string values "all"
        to pick all channels, or "data" to pick data channels. ``None`` (default) will
        pick all MEG and EEG channels, excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent time points
        starting from this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated starting from the first time
        point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and including this
        time point. This value is given in seconds. Defaults to ``None``, in which case
        patches are generated up to and including the last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to use all
        available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need the tqdm
        python module installed. Defaults to False.

    Returns
    -------
    rsa : Evoked | list of Evoked
        The correlation values for each searchlight patch. When spatial_radius is set to
        None, there will only be one virtual sensor. When temporal_radius is set to
        None, there will only be one time point. When multiple models have been
        supplied, a list will be returned containing the RSA results for each model.

    See Also
    --------
    compute_rdm

    """
    one_model = isinstance(rdm_model, list)
    if one_model:
        rdm_model = [rdm_model]

    logger.info(f"Performing RSA between Evokeds and {len(rdm_model)} model RDM(s)")

    # Check for compatibility of the evokeds and the model features
    for rdm in rdm_model:
        n_items = _n_items_from_rdm(rdm)
        if len(evokeds) != n_items and y is None:
            raise ValueError(
                "The number of evokeds (%d) should be equal to the "
                "number of items in `rdm_model` (%d). Alternatively, use "
                "the `y` parameter to assign evokeds to items."
                % (len(evokeds), n_items)
            )
        if y is not None and np.unique(y) != n_items:
            raise ValueError(
                "The number of items in `rdm_model` (%d) does not match "
                "the number of items encoded in the `y` matrix (%d)."
                % (n_items, len(np.unique(y)))
            )

    times = evokeds[0].times
    for evoked in evokeds:
        if np.any(evoked.times != times):
            raise ValueError("Not all evokeds have the same time points.")

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(evokeds[0].info["sfreq"] * temporal_radius)

    # Normalize with the noise cov
    if noise_cov is not None:
        if spatial_radius is not None:
            logger.info(
                "    Using diagonal values of the covariance matrix to whiten "
                "the data."
            )
            diag = True
        else:
            logger.info("    Using covariance matrix to whiten the data.")
            diag = False
        evokeds = [
            mne.whiten_evoked(evoked, noise_cov, diag=diag) for evoked in evokeds
        ]

    # Compute the distances between the sensors
    locs = np.vstack([ch["loc"][:3] for ch in evokeds[0].info["chs"]])
    dist = distance.squareform(distance.pdist(locs))

    picks = mne.io.pick._picks_to_idx(evokeds[0].info, picks, none="data")
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    samples_from, samples_to = _tmin_tmax_to_indices(evokeds[0].times, tmin, tmax)

    if spatial_radius is not None:
        logger.info(f"    Spatial radius: {spatial_radius} meters")
        logger.info(f"    Using {len(picks)} sensors")
    if temporal_radius is not None:
        logger.info(f"    Temporal radius: {temporal_radius} samples")
    if tmin is not None or tmax is not None:
        logger.info(f"    Time interval: {tmin}-{tmax} seconds")

    # Perform the RSA
    X = np.array([evoked.data for evoked in evokeds])
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=temporal_radius,
        sel_series=picks,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    data = rsa_array(
        X,
        rdm_model,
        patches,
        data_rdm_metric=evoked_rdm_metric,
        data_rdm_params=evoked_rdm_params,
        rsa_metric=rsa_metric,
        ignore_nan=ignore_nan,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    # Pack the result in an Evoked object
    if spatial_radius is not None:
        info = mne.pick_info(evokeds[0].info, picks)
    else:
        info = mne.create_info(["rsa"], evokeds[0].info["sfreq"])
    tmin = _construct_tmin(evokeds[0].times, samples_from, samples_to, temporal_radius)

    if one_model:
        return mne.EvokedArray(
            np.atleast_2d(data), info, tmin, comment="RSA", nave=len(evokeds)
        )
    else:
        return [
            mne.EvokedArray(
                np.atleast_2d(data[..., i]),
                info,
                tmin,
                comment="RSA",
                nave=len(evokeds),
            )
            for i in range(data.shape[-1])
        ]


def rsa_epochs(
    epochs,
    rdm_model,
    noise_cov=None,
    spatial_radius=None,
    temporal_radius=None,
    epochs_rdm_metric="correlation",
    epochs_rdm_params=dict(),
    rsa_metric="spearman",
    ignore_nan=False,
    y=None,
    n_folds=1,
    picks=None,
    tmin=None,
    tmax=None,
    dropped_as_nan=False,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA in a searchlight pattern on epochs.

    The output is an Evoked object where the "signal" at each sensor is the RSA,
    computed using all surrounding sensors.

    Parameters
    ----------
    epochs : instance of mne.Epochs
        The brain activity during the epochs. The event codes are used to distinguish
        between items.
    rdm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model RDM, see :func:`compute_rdm`. For efficiency, you can give it in
        condensed form, meaning only the upper triangle of the matrix as a vector. See
        :func:`scipy.spatial.distance.squareform`. To perform RSA against multiple
        models at the same time, supply a list of model RDMs.

        Use :func:`compute_rdm` to compute RDMs.
    noise_cov : mne.Covariance | None
        When specified, the data will by normalized using the noise covariance. This is
        recommended in all cases, but a hard requirement when the data contains sensors
        of different types. Defaults to None.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All sensors within this
        radius will belong to the searchlight patch. Set to None to only perform the
        searchlight over time, flattening across sensors. Defaults to None.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to only
        perform the searchlight over sensors, flattening across time. Defaults to None.
    epochs_rdm_metric : str
        The metric to use to compute the RDM for the epochs. This can be any metric
        supported by the scipy.distance.pdist function. See also the
        ``epochs_rdm_params`` parameter to specify and additional parameter for the
        distance function. Defaults to 'correlation'.
    epochs_rdm_params : dict
        Extra arguments for the distance metric used to compute the RDMs. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing the
        distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    y : ndarray of int, shape (n_items,) | None
        For each Epoch, a number indicating the item to which it belongs. When
        ``None``, the event codes are used to differentiate between items. Defaults to
        ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance metric.
        Folds are created based on the ``y`` parameter. Specify ``None`` to use the
        maximum number of folds possible, given the data. Alternatively, you can pass a
        Scikit-Learn cross validator object (e.g. ``sklearn.model_selection.KFold``) to
        assert fine-grained control over how folds are created.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted as channel
        indices. In lists, channel *type* strings (e.g., ``['meg', 'eeg']``) will pick
        channels of those types, channel *name* strings (e.g., ``['MEG0111',
        'MEG2623']`` will pick the given channels. Can also be the string values "all"
        to pick all channels, or "data" to pick data channels. ``None`` (default) will
        pick all MEG and EEG channels, excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent time points
        starting from this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated starting from the first time
        point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and including this
        time point. This value is given in seconds. Defaults to ``None``, in which case
        patches are generated up to and including the last time point.
    dropped_as_nan : bool
        When this is set to ``True``, the drop log will be used to inject NaN values in
        the RDMs at the locations where a bad epoch was dropped. This is useful to
        ensure the dimensions of the RDM are the same, irregardless of any bad epochs
        that were dropped. Make sure to use ``ignore_nan=True`` when using RDMs with
        NaNs in them during subsequent RSA computations. Defaults to ``False``.

        .. versionadded:: 0.8
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to use all
        available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need the tqdm
        python module installed. Defaults to False.

    Returns
    -------
    rsa : Evoked | list of Evoked
        The correlation values for each searchlight patch. When spatial_radius is set to
        None, there will only be one virtual sensor. When temporal_radius is set to
        None, there will only be one time point. When multiple models have been
        supplied, a list will be returned containing the RSA results for each model.

    See Also
    --------
    compute_rdm

    """
    one_model = type(rdm_model) is np.ndarray
    if one_model:
        rdm_model = [rdm_model]

    logger.info(f"Performing RSA between Epochs and {len(rdm_model)} model RDM(s)")

    if y is None:
        y_source = "Epoch object"
        y = epochs.events[:, 2]
    else:
        y_source = "`y` matrix"

    # Check for compatibility of the epochs and the model features
    for rdm in rdm_model:
        n_items = _n_items_from_rdm(rdm)
        if len(np.unique(y)) != n_items:
            raise ValueError(
                "The number of items in `rdm_model` (%d) does not match "
                "the number of items encoded in the %s (%d)."
                % (n_items, y_source, len(np.unique(y)))
            )

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(epochs.info["sfreq"] * temporal_radius)

    picks = mne.io.pick._picks_to_idx(epochs.info, picks, none="data")
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    samples_from, samples_to = _tmin_tmax_to_indices(epochs.times, tmin, tmax)

    # Normalize with the noise cov
    if noise_cov is not None:
        if spatial_radius is not None:
            logger.info(
                "    Using diagonal values of the covariance matrix to whiten "
                "the data."
            )
            noise_cov = noise_cov.as_diag()
        else:
            logger.info("    Using covariance matrix to whiten the data.")
        W, _ = compute_whitener(noise_cov, epochs.info, picks=picks)
        epochs._data[:, picks] = (W @ epochs._data[:, picks].T).T

    if spatial_radius is not None:
        logger.info(f"    Spatial radius: {spatial_radius} meters")
        logger.info(f"    Using {len(picks)} sensors")

        # Compute the distances between the sensors
        locs = np.vstack([ch["loc"][:3] for ch in epochs.info["chs"]])
        dist = distance.squareform(distance.pdist(locs))
    else:
        dist = None

    if temporal_radius is not None:
        logger.info(f"    Temporal radius: {temporal_radius} samples")
        logger.info(f"    Time interval: {tmin}-{tmax} seconds")

    # Perform the RSA
    X = epochs.get_data(copy=False)
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=temporal_radius,
        sel_series=picks,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    data = rsa_array(
        X,
        rdm_model,
        patches,
        data_rdm_metric=epochs_rdm_metric,
        data_rdm_params=epochs_rdm_params,
        rsa_metric=rsa_metric,
        ignore_nan=ignore_nan,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    # Pack the result in an Evoked object
    if spatial_radius is not None:
        info = epochs.info
        info = mne.pick_info(info, picks)
    else:
        info = mne.create_info(["rsa"], epochs.info["sfreq"])
    tmin = _construct_tmin(epochs.times, samples_from, samples_to, temporal_radius)

    if one_model:
        return mne.EvokedArray(
            np.atleast_2d(data), info, tmin, comment="RSA", nave=len(np.unique(y))
        )
    else:
        return [
            mne.EvokedArray(
                np.atleast_2d(data[..., i]),
                info,
                tmin,
                comment="RSA",
                nave=len(np.unique(y)),
            )
            for i in range(data.shape[-1])
        ]


def rdm_evokeds(
    evokeds,
    noise_cov=None,
    spatial_radius=None,
    temporal_radius=None,
    dist_metric="correlation",
    dist_params=dict(),
    y=None,
    n_folds=1,
    picks=None,
    tmin=None,
    tmax=None,
):
    """Generate RDMs in a searchlight pattern on evokeds.

    Parameters
    ----------
    evokeds : list of mne.Evoked
        The evoked brain activity for each item. If you have more than one Evoked object
        per item (i.e. repetitions), you can use the ``y`` parameter to assign evokeds
        to items.
    noise_cov : mne.Covariance | None
        When specified, the data will by normalized using the noise covariance. This is
        recommended in all cases, but a hard requirement when the data contains sensors
        of different types. Defaults to None.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All sensors within this
        radius will belong to the searchlight patch. Set to None to only perform the
        searchlight over time, flattening across sensors. Defaults to None.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to only
        perform the searchlight over sensors, flattening across time. Defaults to None.
    dist_metric : str
        The metric to use to compute the RDM for the evokeds. This can be any metric
        supported by the scipy.distance.pdist function. See also the ``dist_params``
        parameter to specify and additional parameter for the distance function.
        Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the RDMs. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each Evoked, a number indicating the item to which it belongs. When
        ``None``, each Evoked is assumed to belong to a different item.
        Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Alternatively, you can pass a Scikit-Learn cross validator object (e.g.
        ``sklearn.model_selection.KFold``) to assert fine-grained control over
        how folds are created.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted as channel
        indices. In lists, channel *type* strings (e.g., ``['meg', 'eeg']``) will pick
        channels of those types, channel *name* strings (e.g., ``['MEG0111',
        'MEG2623']`` will pick the given channels. Can also be the string values "all"
        to pick all channels, or "data" to pick data channels. ``None`` (default) will
        pick all MEG and EEG channels, excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent time points
        starting from this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated starting from the first time
        point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and including this
        time point. This value is given in seconds. Defaults to ``None``, in which case
        patches are generated up to and including the last time point.

    Yields
    ------
    rdm : ndarray, shape (n_items, n_items)
        A RDM for each searchlight patch.

    """
    times = evokeds[0].times
    for evoked in evokeds:
        if np.any(evoked.times != times):
            raise ValueError("Not all evokeds have the same time points.")

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(evokeds[0].info["sfreq"] * temporal_radius)

    # Normalize with the noise cov
    if noise_cov is not None:
        diag = spatial_radius is not None
        evokeds = [
            mne.whiten_evoked(evoked, noise_cov, diag=diag) for evoked in evokeds
        ]

    # Compute the distances between the sensors
    locs = np.vstack([ch["loc"][:3] for ch in evokeds[0].info["chs"]])
    dist = distance.squareform(distance.pdist(locs))

    picks = mne.io.pick._picks_to_idx(evokeds[0].info, picks, none="data")
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    samples_from, samples_to = _tmin_tmax_to_indices(times, tmin, tmax)

    # Compute the RDMs
    X = np.array([evoked.data for evoked in evokeds])
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=temporal_radius,
        sel_series=picks,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    yield from rdm_array(
        X,
        patches,
        dist_metric=dist_metric,
        dist_params=dist_params,
        y=y,
        n_folds=n_folds,
    )


def rdm_epochs(
    epochs,
    noise_cov=None,
    spatial_radius=None,
    temporal_radius=None,
    dist_metric="correlation",
    dist_params=dict(),
    y=None,
    n_folds=1,
    picks=None,
    tmin=None,
    tmax=None,
    dropped_as_nan=False,
):
    """Generate RDMs in a searchlight pattern on epochs.

    Parameters
    ----------
    epochs : instance of mne.Epochs
        The brain activity during the epochs. The event codes are used to distinguish
        between items.
    noise_cov : mne.Covariance | None
        When specified, the data will by normalized using the noise covariance. This is
        recommended in all cases, but a hard requirement when the data contains sensors
        of different types. Defaults to None.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All sensors within this
        radius will belong to the searchlight patch. Set to None to only perform the
        searchlight over time, flattening across sensors. Defaults to None.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to only
        perform the searchlight over sensors, flattening across time. Defaults to None.
    dist_metric : str
        The metric to use to compute the RDM for the epochs. This can be any metric
        supported by the scipy.distance.pdist function. See also the
        ``epochs_rdm_params`` parameter to specify and additional parameter for the
        distance function. Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the RDMs. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each Epoch, a number indicating the item to which it belongs. When
        ``None``, the event codes are used to differentiate between items. Defaults to
        ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance metric.
        Folds are created based on the ``y`` parameter. Specify ``None`` to use the
        maximum number of folds possible, given the data. Alternatively, you can pass a
        Scikit-Learn cross validator object (e.g. ``sklearn.model_selection.KFold``) to
        assert fine-grained control over how folds are created.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted as channel
        indices. In lists, channel *type* strings (e.g., ``['meg', 'eeg']``) will pick
        channels of those types, channel *name* strings (e.g., ``['MEG0111',
        'MEG2623']`` will pick the given channels. Can also be the string values "all"
        to pick all channels, or "data" to pick data channels. ``None`` (default) will
        pick all MEG and EEG channels, excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent time points
        starting from this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated starting from the first time
        point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and including this
        time point. This value is given in seconds. Defaults to ``None``, in which case
        patches are generated up to and including the last time point.
    dropped_as_nan : bool
        When this is set to ``True``, the drop log will be used to inject NaN values in
        the RDMs at the locations where a bad epoch was dropped. This is useful to
        ensure the dimensions of the RDM are the same, irregardless of any bad epochs
        that were dropped. Make sure to use ``ignore_nan=True`` when using RDMs with
        NaNs in them during subsequent RSA computations. Defaults to ``False``.

        .. versionadded:: 0.8

    Yields
    ------
    rdm : ndarray, shape (n_items, n_items)
        A RDM for each searchlight patch.

    """
    if y is None:
        y = epochs.events[:, 2]

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(epochs.info["sfreq"] * temporal_radius)

    picks = mne.io.pick._picks_to_idx(epochs.info, picks, none="data")
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    samples_from, samples_to = _tmin_tmax_to_indices(epochs.times, tmin, tmax)

    # Normalize with the noise cov
    if noise_cov is not None:
        if spatial_radius is not None:
            logger.info(
                "    Using diagonal values of the covariance matrix to whiten "
                "the data."
            )
            noise_cov = noise_cov.as_diag()
        else:
            logger.info("    Using covariance matrix to whiten the data.")
        W, _ = compute_whitener(noise_cov, epochs.info, picks=picks)
        epochs._data[picks] = W @ epochs._data[picks]

    if spatial_radius is not None:
        # Compute the distances between the sensors
        locs = np.vstack([ch["loc"][:3] for ch in epochs.info["chs"]])
        dist = distance.squareform(distance.pdist(locs))
    else:
        dist = None

    # Compute the RDMs
    X = epochs.get_data(copy=False)
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=temporal_radius,
        sel_series=picks,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    rdm_gen = rdm_array(
        X,
        patches,
        dist_metric=dist_metric,
        dist_params=dist_params,
        y=y,
        n_folds=n_folds,
    )
    if not dropped_as_nan or epochs.drop_log_stats() == 0:
        yield from rdm_gen
    else:
        nan_locations = [
            i for i, reason in enumerate(epochs.drop_log) if len(reason) > 0
        ]
        for rdm in rdm_gen:
            rdm = distance.squareform(rdm)
            rdm = np.insert(rdm, nan_locations, np.NaN, axis=0)
            rdm = np.insert(rdm, nan_locations, np.NaN, axis=1)
            # Can't use squareform to convert back due to the NaNs.
            yield rdm[np.triu_indices(len(rdm), 1)]


def _tmin_tmax_to_indices(times, tmin, tmax):
    """Convert tmin tmax parameters to an array of sample indices."""
    if tmin is None:
        samples_from = 0
    else:
        samples_from = np.searchsorted(times, tmin)
    if tmax is None:
        samples_to = len(times)
    else:
        samples_to = np.searchsorted(times, tmax)
    if samples_from > samples_to:
        raise ValueError(f"Invalid time range: {tmin} to {tmax}")
    return samples_from, samples_to


def _construct_tmin(times, samples_from, samples_to, temporal_radius):
    if temporal_radius is None:
        return times[(samples_from + samples_to) // 2]
    else:
        return times[max(temporal_radius, samples_from)]


def _square_to_condensed(i, j, n):
    if i < j:
        i, j = j, i
    return n * j - j * (j + 1) // 2 + i - 1 - j



"""Unit tests for comparing RDMs (which is called RSA)."""

from types import GeneratorType

import numpy as np
import pytest
from mne_rsa import rsa, rsa_array, rsa_gen, searchlight
from mne_rsa.rsa import _kendall_tau_a, _partial_correlation
from numpy.testing import assert_allclose, assert_equal


def rdm():
    """Create an example RDM."""
    return np.array([1, 2, 3, 4, 5, 6])


def rdm_gen(rdms):
    """Generate RDMs."""
    for rdm in rdms:
        yield np.asarray(rdm)


class TestRSAGen:
    """Test the rsa_gen function."""

    def test_return_type(self):
        """Test return type of rsa_gen."""
        assert isinstance(rsa_gen(rdm_gen([rdm()]), rdm()), GeneratorType)
        assert next(rsa_gen(rdm_gen([rdm()]), rdm())).shape == tuple()
        assert next(rsa_gen(rdm_gen([rdm()]), [rdm()])).shape == (1,)

    def test_spearman(self):
        """Test computing RSA with Spearman correlation."""
        data_rdm = rdm_gen([[1, 2, 3]])
        model_rdm = np.array([2, 3, 3.5])
        assert next(rsa_gen(data_rdm, model_rdm, metric="spearman")) == 1.0

        data_rdm = rdm_gen([[np.NaN, 2, 3, 4, 5]])
        model_rdm = np.array([2, np.NaN, 3.5, 4, 5])
        assert (
            next(rsa_gen(data_rdm, model_rdm, metric="spearman", ignore_nan=True))
            == 1.0
        )

    def test_pearson(self):
        """Test computing RSA with Pearson correlation."""
        data_rdm = rdm_gen([[1, 2, 3]])
        model_rdm = np.array([2, 3, 3.5])
        assert next(rsa_gen(data_rdm, model_rdm, metric="pearson")) < 1.0

        data_rdm = rdm_gen([[np.NaN, 2, 3, 4, 5]])
        model_rdm = np.array([2, np.NaN, 3.5, 4, 5])
        assert (
            next(rsa_gen(data_rdm, model_rdm, metric="pearson", ignore_nan=True)) < 1.0
        )

    def test_kendall_tau_a(self):
        """Test computing RSA with Kendall's Tau Alpha."""
        data_rdm = rdm_gen([[1, 2, 3]])
        model_rdm = np.array([1, 3, 3])  # This metric deals well with ties
        rsa_val = next(rsa_gen(data_rdm, model_rdm, metric="kendall-tau-a"))
        assert rsa_val == 2 / 3

        data_rdm = rdm_gen([[1, np.NaN, 2, 3, 4]])
        model_rdm = np.array([1, 2, 3, 3, np.NaN])
        rsa_val = next(
            rsa_gen(data_rdm, model_rdm, metric="kendall-tau-a", ignore_nan=True)
        )
        assert rsa_val == 2 / 3

    def test_regression(self):
        """Test computing RSA with regression."""
        model_rdm1 = np.array([-1, 0, 1])
        model_rdm2 = np.array([1, -2, 1])
        data_rdm = rdm_gen([3 * model_rdm1 + 5 * model_rdm2])
        rsa_val = next(rsa_gen(data_rdm, [model_rdm1, model_rdm2], metric="regression"))
        assert_allclose(rsa_val, [3, 5])

        model_rdm1 = np.array([-1, np.NaN, 0, 1, 1])
        model_rdm2 = np.array([1, 3, -2, 1, np.NaN])
        data_rdm = rdm_gen([3 * model_rdm1 + 5 * model_rdm2])
        rsa_val = next(
            rsa_gen(
                data_rdm, [model_rdm1, model_rdm2], metric="regression", ignore_nan=True
            )
        )
        assert_allclose(rsa_val, [3, 5])

    def test_partial(self):
        """Test computing RSA with partial correlation."""
        # Example taken from https://en.wikipedia.org/wiki/Partial_correlation
        model_rdm1 = np.array([1, 2, 3, 4])
        model_rdm2 = np.array([0, 0, 1, 1])
        data_rdm = rdm_gen([[2, 4, 15, 20]])
        rsa_val = next(rsa_gen(data_rdm, [model_rdm1, model_rdm2], metric="partial"))
        assert_allclose(rsa_val, [0.919145, 0.912871])

        model_rdm1 = np.array([1, np.NaN, 2, 3, 4, 4])
        model_rdm2 = np.array([0, 0, 0, 1, 1, np.NaN])
        data_rdm = rdm_gen([[2, np.NaN, 4, 15, 20, np.NaN]])
        rsa_val = next(
            rsa_gen(
                data_rdm, [model_rdm1, model_rdm2], metric="partial", ignore_nan=True
            )
        )
        assert_allclose(rsa_val, [0.919145, 0.912871])

    def test_partial_spearman(self):
        """Test computing RSA with partial spearman correlation."""
        # Example verified with MATLAB's partialcorr function
        model_rdm1 = np.array([1, 2, 3, 4])
        model_rdm2 = np.array([0, 0, 1, 1])
        data_rdm = rdm_gen([[2, 4, 20, 15]])
        rsa_val = next(
            rsa_gen(data_rdm, [model_rdm1, model_rdm2], metric="partial-spearman")
        )
        assert_allclose(rsa_val, [0, 2 / 3], atol=1e-15)

        model_rdm1 = np.array([1, np.NaN, 2, 3, 4, 4])
        model_rdm2 = np.array([0, 0, 0, 1, 1, np.NaN])
        data_rdm = rdm_gen([[2, np.NaN, 4, 20, 15, np.NaN]])
        rsa_val = next(
            rsa_gen(
                data_rdm,
                [model_rdm1, model_rdm2],
                metric="partial-spearman",
                ignore_nan=True,
            )
        )
        assert_allclose(rsa_val, [0, 2 / 3], atol=1e-15)

    def test_invalid_metric(self):
        """Test whether computing RSA with an invalid metric raises error."""
        with pytest.raises(ValueError, match="Invalid RSA metric"):
            next(rsa_gen(rdm_gen([rdm()]), rdm(), metric="foo"))

        # These metrics only work with multiple model RDMs
        with pytest.raises(ValueError, match="Need more than one model RDM"):
            next(rsa_gen(rdm_gen([rdm()]), rdm(), metric="partial"))
        with pytest.raises(ValueError, match="Need more than one model RDM"):
            next(rsa_gen(rdm_gen([rdm()]), rdm(), metric="partial-spearman"))

    def test_nan(self):
        """Test whether NaNs generate an error when appropriate."""
        assert np.isnan(next(rsa_gen(rdm_gen([[1, 2, np.NaN, 4, 5, 6]]), rdm())))
        assert_allclose(
            next(rsa_gen(rdm_gen([[1, 2, np.NaN, 4, 5, 6]]), rdm(), ignore_nan=True)),
            1,
            atol=1e-15,
        )


class TestRSA:
    """Test the main RSA function."""

    # Most of the functionality is already tested in TestRSAGen
    def test_return_type(self):
        """Test return type of rsa_gen."""
        assert isinstance(rsa([rdm()], rdm()), np.ndarray)
        assert rsa(rdm(), rdm()).shape == tuple()
        assert rsa(rdm(), [rdm()]).shape == (1,)

    def test_progress_bar(self):
        """Test showing a progress bar for rsa."""
        assert rsa([rdm()], rdm(), verbose=True) == [1.0]
        assert rsa([rdm()], rdm(), verbose=True, n_data_rdms=1) == [1.0]
        assert rsa(rdm_gen([rdm()]), rdm(), verbose=True) == [1.0]

    def test_parallel_processing(self):
        """Test running rsa with multiple cores."""
        assert_equal(rsa([rdm(), rdm()], rdm(), n_jobs=2), [1.0, 1.0])


class TestRSASearchlight:
    """Test computing RSA using searchlight patches."""

    def test_invalid_input(self):
        """Test invalid inputs."""
        data = np.array([[1], [2], [3], [4]])
        model_rdm = rdm()
        with pytest.raises(ValueError, match="There is only a single feature"):
            rsa_array(data, model_rdm)

    def test_rsa_single_searchlight_patch(self):
        """Test RSA with a single searchlight patch."""
        data = np.array([[1], [2], [3], [4]])
        model_rdm = np.array([1, 2, 3, 1, 2, 1])
        rsa_result = rsa_array(data, model_rdm, data_rdm_metric="euclidean")
        assert rsa_result.shape == tuple()  # Single scalar value
        assert rsa_result == 1

    def test_rsa_temp(self):
        """Test RSA with a temporal searchlight."""
        data = np.array([[1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6]])
        model_rdm = np.array([1, 2, 1])

        # One model RDM, no parallelization
        patches = searchlight(data.shape, temporal_radius=1)
        rsa_result = rsa_array(data, model_rdm, patches, data_rdm_metric="euclidean")
        assert rsa_result.shape == (2,)
        assert_equal(rsa_result, 1)

        # Multiple model RDMs, no parallelization
        patches = searchlight(data.shape, temporal_radius=1)
        rsa_result = rsa_array(
            data, [model_rdm, model_rdm], patches, data_rdm_metric="euclidean"
        )
        assert rsa_result.shape == (2, 2)
        assert_equal(rsa_result, 1)

        # One model RDM, parallelization across 2 CPUs
        patches = searchlight(data.shape, temporal_radius=1)
        rsa_result = rsa_array(
            data, model_rdm, patches, data_rdm_metric="euclidean", n_jobs=2
        )
        assert rsa_result.shape == (2,)
        assert_equal(rsa_result, 1)

        # Multiple model RDMs, parallelization across 2 CPUs
        patches = searchlight(data.shape, temporal_radius=1)
        rsa_result = rsa_array(
            data, [model_rdm, model_rdm], patches, data_rdm_metric="euclidean", n_jobs=2
        )
        assert rsa_result.shape == (2, 2)
        assert_equal(rsa_result, 1)

    def test_rsa_spat(self):
        """Test RSA with a spatial searchlight."""
        data = np.array([[[1], [2], [3]], [[2], [3], [4]], [[3], [4], [5]]])
        model_rdm = np.array([1, 2, 1])
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])

        # One model RDM, no paralellization
        patches = searchlight(data.shape, dist, spatial_radius=1)
        rsa_result = rsa_array(data, model_rdm, patches, data_rdm_metric="euclidean")
        assert rsa_result.shape == (3,)
        assert_equal(rsa_result, 1)

        # Multiple model RDMs, no paralellization
        patches = searchlight(data.shape, dist, spatial_radius=1)
        rsa_result = rsa_array(
            data, [model_rdm, model_rdm], patches, data_rdm_metric="euclidean"
        )
        assert rsa_result.shape == (3, 2)
        assert_equal(rsa_result, 1)

        # One model RDM, paralellization across 2 CPUs
        patches = searchlight(data.shape, dist, spatial_radius=1)
        rsa_result = rsa_array(
            data, model_rdm, patches, data_rdm_metric="euclidean", n_jobs=2
        )
        assert rsa_result.shape == (3,)
        assert_equal(rsa_result, 1)

        # Multiple model RDMs, paralellization across 2 CPUs
        patches = searchlight(data.shape, dist, spatial_radius=1)
        rsa_result = rsa_array(
            data, [model_rdm, model_rdm], patches, data_rdm_metric="euclidean", n_jobs=2
        )
        assert rsa_result.shape == (3, 2)
        assert_equal(rsa_result, 1)

    def test_rsa_spat_temp(self):
        """Test RSA with a spatial-temporal searchlight."""
        data = np.array(
            [[[1, 2, 3], [2, 3, 4]], [[2, 3, 4], [3, 4, 5]], [[3, 4, 5], [4, 5, 6]]]
        )
        model_rdm = np.array([1, 2, 1])
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])

        # One model RDM, no paralellization
        patches = searchlight(data.shape, dist, spatial_radius=1, temporal_radius=1)
        rsa_result = rsa_array(data, model_rdm, patches, data_rdm_metric="euclidean")
        assert rsa_result.shape == (2, 1)
        assert_equal(rsa_result, 1)

        # Multiple model RDMs, no paralellization
        patches = searchlight(data.shape, dist, spatial_radius=1, temporal_radius=1)
        rsa_result = rsa_array(
            data, [model_rdm, model_rdm], patches, data_rdm_metric="euclidean"
        )
        assert rsa_result.shape == (2, 1, 2)
        assert_equal(rsa_result, 1)

        # One model RDM, paralellization across 2 CPUs
        patches = searchlight(data.shape, dist, spatial_radius=1, temporal_radius=1)
        rsa_result = rsa_array(
            data, model_rdm, patches, data_rdm_metric="euclidean", n_jobs=2
        )
        assert rsa_result.shape == (2, 1)
        assert_equal(rsa_result, 1)

        # Multiple model RDMs, paralellization across 2 CPUs
        patches = searchlight(data.shape, dist, spatial_radius=1, temporal_radius=1)
        rsa_result = rsa_array(
            data, [model_rdm, model_rdm], patches, data_rdm_metric="euclidean", n_jobs=2
        )
        assert rsa_result.shape == (2, 1, 2)
        assert_equal(rsa_result, 1)


class TestKendallTau:
    """Test computing Kendall's Tau Alpha metric."""

    def test_basic(self):
        """Test computing Kendall's Tau Alpha."""
        # This metric deals well with ties
        assert _kendall_tau_a([1, 2, 3], [1, 3, 3]) == 2 / 3

        # Test taken from scipy
        assert _kendall_tau_a([9, 2, 5, 6], [4, 7, 9, 11]) == 0

    def test_sizes(self):
        """Test feeding arrays to Kendall's Tau Alpha of unequal size."""
        with pytest.raises(ValueError, match="must be of the same size"):
            _kendall_tau_a([1, 2, 3], [1, 2, 3, 4, 5])

    def test_empty(self):
        """Test feeding empty arrays to Kendall's Tau Alpha."""
        assert np.isnan(_kendall_tau_a([], []))

    def test_only_ties(self):
        """Test Kendall's Tau Alpha where every value is a tie."""
        assert np.isnan(_kendall_tau_a([1, 1, 1], [2, 2, 2]))


class TestPartialCorrelation:
    """Test partial correlations."""

    def test_basic(self):
        """Test computing partial correlations."""
        data_rdm = [2, 4, 15, 20]
        model_rdms = [[1, 2, 3, 4], [0, 0, 1, 1]]
        assert_allclose(
            _partial_correlation(data_rdm, model_rdms, type="pearson"),
            [0.919145, 0.912871],
        )
        print(_partial_correlation(data_rdm, model_rdms, type="spearman"))
        assert_allclose(
            _partial_correlation(data_rdm, model_rdms, type="spearman"),
            [-1, 0.89442719],
        )

    def test_sizes(self):
        """Test feeding arrays of invalid size to _partial_correlation."""
        with pytest.raises(ValueError, match="Need more than one model RDM"):
            _partial_correlation([1, 2], [[1, 2]])
        with pytest.raises(ValueError, match="Correlation type must be"):
            _partial_correlation([1, 2], [[1, 2], [1, 2]], type="banana")



# encoding: utf-8
"""Module implementing representational similarity analysis (RSA) at the source level.

Kriegeskorte, N., Mur, M., & Bandettini, P. A. (2008). Representational
similarity analysis - connecting the branches of systems neuroscience.
Frontiers in Systems Neuroscience, 2(November), 4.
https://doi.org/10.3389/neuro.06.004.2008

Authors
-------
Marijn van Vliet <marijn.vanvliet@aalto.fi>
Annika Hultén <annika.hulten@aalto.fi>
Ossi Lehtonen <ossi.lehtonen@aalto.fi>
"""

from copy import deepcopy
from warnings import warn
import numpy as np
import mne
from mne.utils import logger, verbose
from scipy.linalg import block_diag
import nibabel as nib

from .rdm import _n_items_from_rdm, rdm_array
from .rsa import rsa_array
from .searchlight import searchlight
from .sensor_level import _tmin_tmax_to_indices, _construct_tmin


@verbose
def rsa_stcs(
    stcs,
    rdm_model,
    src,
    spatial_radius=None,
    temporal_radius=None,
    stc_rdm_metric="correlation",
    stc_rdm_params=dict(),
    rsa_metric="spearman",
    ignore_nan=False,
    y=None,
    n_folds=1,
    sel_vertices=None,
    sel_vertices_by_index=None,
    tmin=None,
    tmax=None,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA in a searchlight pattern on MNE-Python source estimates.

    The output is a source estimate where the "signal" at each source point is
    the RSA, computed for a patch surrounding the source point. Source estimate
    objects can be either defined along a cortical surface (``SourceEstimate``
    objects) or volumetric (``VolSourceEstimate`` objects).  For surface source
    estimates, distances between vertices are measured in 2D space, namely as
    the length of the path along the surface from one vertex to another. For
    volume source estimates, distances are measured in 3D space as a straight
    line from one voxel to another.

    Parameters
    ----------
    stcs : list of mne.SourceEstimate | list of mne.VolSourceEstimate
        For each item, a source estimate for the brain activity.
    rdm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model RDM, see :func:`compute_rdm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model RDMs.

        Use :func:`compute_rdm` to compute RDMs.
    src : instance of mne.SourceSpaces
        The source space used by the source estimates specified in the `stcs`
        parameter.
    spatial_radius : float | None
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch. Set to
        None to only perform the searchlight over time, flattening across
        sensors. Defaults to None.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to None.
    stc_rdm_metric : str
        The metric to use to compute the RDM for the source estimates. This can
        be any metric supported by the scipy.distance.pdist function. See also
        the ``stc_rdm_params`` parameter to specify and additional parameter
        for the distance function. Defaults to 'correlation'.
    stc_rdm_params : dict
        Extra arguments for the distance metric used to compute the RDMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing
        the distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Alternatively, you can pass a Scikit-Learn cross validator object (e.g.
        ``sklearn.model_selection.KFold``) to assert fine-grained control over
        how folds are created.
        Defaults to 1 (no cross-validation).
    sel_vertices :  mne.Label | list of mne.Label | list of ndarray | None
        When set, searchlight patches will only be generated for the subset of
        ROI labels, or vertices/voxels with the given vertex numbers. When
        giving vertex numbers, supply a list of numpy arrays with for each
        hemisphere, the selected vertex numbers. For volume source spaces,
        supple a list with only a single element, with that element being the
        ndarray with vertex numbers.
        See also ``sel_vertices_by_index`` for an alternative manner of
        selecting vertices.
    sel_vertices_by_index : ndarray of int, shape (n_vertices,)
        When set, searchlight patches will only be generated for the subset of
        vertices with the given indices in the ``stc.data`` array.
        See also ``sel_vertices`` for an alternative manner of selecting vertices.
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    stc : SourceEstimate | VolSourceEstimate | list of SourceEstimate | list of VolSourceEstimate
        The correlation values for each searchlight patch. When spatial_radius
        is set to None, there will only be one vertex. When temporal_radius is
        set to None, there will only be one time point. When multiple models
        have been supplied, a list will be returned containing the RSA results
        for each model.

    See Also
    --------
    compute_rdm
    """  # noqa E501
    # Check for compatibility of the source estimates and the model features
    one_model = type(rdm_model) is np.ndarray
    if one_model:
        rdm_model = [rdm_model]

    # Check for compatibility of the stcs and the model features
    for rdm in rdm_model:
        n_items = _n_items_from_rdm(rdm)
        if len(stcs) != n_items and y is None:
            raise ValueError(
                "The number of source estimates (%d) should be equal to the "
                "number of items in `rdm_model` (%d). Alternatively, use "
                "the `y` parameter to assign source estimates to items."
                % (len(stcs), n_items)
            )
        if y is not None and len(np.unique(y)) != n_items:
            raise ValueError(
                "The number of items in `rdm_model` (%d) does not match "
                "the number of items encoded in the `y` matrix (%d)."
                % (n_items, len(np.unique(y)))
            )

    src = _check_stcs_compatibility(stcs, src)
    if spatial_radius is not None:
        dist = _get_distance_matrix(src, dist_lim=spatial_radius, n_jobs=n_jobs)
    else:
        dist = None
    if temporal_radius is not None:
        # Convert the temporal radius to samples
        temporal_radius = int(temporal_radius // stcs[0].tstep)
        if temporal_radius < 1:
            raise ValueError("Temporal radius is less than one sample.")

    samples_from, samples_to = _tmin_tmax_to_indices(stcs[0].times, tmin, tmax)

    if sel_vertices_by_index is not None:
        sel_series = sel_vertices_by_index
    elif sel_vertices is None:
        sel_series = np.arange(len(stcs[0].data))
    else:
        sel_series = vertex_selection_to_indices(stcs[0].vertices, sel_vertices)

    logger.info(
        f"Performing RSA between SourceEstimates and {len(rdm_model)} model RDM(s)"
    )
    if spatial_radius is not None:
        logger.info(f"    Spatial radius: {spatial_radius} meters")
    if sel_vertices is not None:
        logger.info(f"    Using {len(sel_series)} vertices")
    else:
        logger.info(f"    Using {sum(len(v) for v in stcs[0].vertices)} vertices")
    if temporal_radius is not None:
        logger.info(f"    Temporal radius: {temporal_radius} samples")
    if tmin is not None or tmax is not None:
        logger.info(f"    Time interval: {tmin}-{tmax} seconds")

    # Perform the RSA
    X = np.array([stc.data for stc in stcs])
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=temporal_radius,
        sel_series=sel_series,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    logger.info(f"    Number of searchlight patches: {len(patches)}")

    data = rsa_array(
        X,
        rdm_model,
        patches,
        data_rdm_metric=stc_rdm_metric,
        data_rdm_params=stc_rdm_params,
        rsa_metric=rsa_metric,
        ignore_nan=ignore_nan,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    # Pack the result in a SourceEstimate object
    if spatial_radius is not None:
        vertices = vertex_indices_to_numbers(stcs[0].vertices, sel_series)
    else:
        if src.kind == "volume":
            vertices = [np.array([1])]
        else:
            vertices = [np.array([1]), np.array([])]
        data = data[np.newaxis, ...]
    tmin = _construct_tmin(stcs[0].times, samples_from, samples_to, temporal_radius)
    tstep = stcs[0].tstep

    if one_model:
        if src.kind == "volume":
            return mne.VolSourceEstimate(
                data, vertices, tmin, tstep, subject=stcs[0].subject
            )
        else:
            return mne.SourceEstimate(
                data, vertices, tmin, tstep, subject=stcs[0].subject
            )
    else:
        if src.kind == "volume":
            return [
                mne.VolSourceEstimate(
                    data[..., i], vertices, tmin, tstep, subject=stcs[0].subject
                )
                for i in range(data.shape[-1])
            ]
        else:
            return [
                mne.SourceEstimate(
                    data[..., i], vertices, tmin, tstep, subject=stcs[0].subject
                )
                for i in range(data.shape[-1])
            ]


@verbose
def rdm_stcs(
    stcs,
    src,
    spatial_radius=None,
    temporal_radius=None,
    dist_metric="sqeuclidean",
    dist_params=dict(),
    y=None,
    n_folds=None,
    sel_vertices=None,
    sel_vertices_by_index=None,
    tmin=None,
    tmax=None,
    n_jobs=1,
    verbose=False,
):
    """Generate RDMs in a searchlight pattern on MNE-Python source estimates.

    RDMs are computed using a patch surrounding each source point. Source
    estimate objects can be either defined along a cortical surface
    (``SourceEstimate`` objects) or volumetric (``VolSourceEstimate`` objects).
    For surface source estimates, distances between vertices are measured in 2D
    space, namely as the length of the path along the surface from one vertex
    to another. For volume source estimates, distances are measured in 3D space
    as a straight line from one voxel to another.

    Parameters
    ----------
    stcs : list of mne.SourceEstimate | list of mne.VolSourceEstimate
        For each item, a source estimate for the brain activity.
    src : instance of mne.SourceSpaces
        The source space used by the source estimates specified in the `stcs`
        parameter.
    spatial_radius : float | None
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch. Set to
        None to only perform the searchlight over time, flattening across
        sensors. Defaults to None.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to None.
    dist_metric : str
        The metric to use to compute the RDM for the source estimates. This can
        be any metric supported by the scipy.distance.pdist function. See also
        the ``stc_rdm_params`` parameter to specify and additional parameter
        for the distance function. Defaults to 'sqeuclidean'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the RDMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Alternatively, you can pass a Scikit-Learn cross validator object (e.g.
        ``sklearn.model_selection.KFold``) to assert fine-grained control over
        how folds are created.
        Defaults to 1 (no cross-validation).
    sel_vertices :  mne.Label | list of mne.Label | list of ndarray | None
        When set, searchlight patches will only be generated for the subset of
        ROI labels, or vertices/voxels with the given vertex numbers. When
        giving vertex numbers, supply a list of numpy arrays with for each
        hemisphere, the selected vertex numbers. For volume source spaces,
        supple a list with only a single element, with that element being the
        ndarray with vertex numbers.
        See also ``sel_vertices_by_index`` for an alternative manner of
        selecting vertices.
    sel_vertices_by_index : ndarray of int, shape (n_vertices,)
        When set, searchlight patches will only be generated for the subset of
        vertices with the given indices in the ``stc.data`` array.
        See also ``sel_vertices`` for an alternative manner of selecting vertices.
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use for the
        source-to-source distance computation. Specify -1 to use all available
        cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Yields
    ------
    rdm : ndarray, shape (n_items, n_items)
        A RDM for each searchlight patch.

    """
    src = _check_stcs_compatibility(stcs, src)
    if spatial_radius is not None:
        dist = _get_distance_matrix(src, dist_lim=spatial_radius, n_jobs=n_jobs)
    else:
        dist = None

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = int(temporal_radius // stcs[0].tstep)

        if temporal_radius < 1:
            raise ValueError("Temporal radius is less than one sample.")

    samples_from, samples_to = _tmin_tmax_to_indices(stcs[0].times, tmin, tmax)

    if sel_vertices_by_index is not None:
        sel_series = sel_vertices_by_index
    elif sel_vertices is None:
        sel_series = np.arange(len(stcs[0].data))
    else:
        sel_series = vertex_selection_to_indices(stcs[0].vertices, sel_vertices)

    X = np.array([stc.data for stc in stcs])
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=temporal_radius,
        sel_series=sel_series,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    yield from rdm_array(
        X,
        patches,
        dist_metric=dist_metric,
        dist_params=dist_params,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
    )


@verbose
def rsa_stcs_rois(
    stcs,
    rdm_model,
    src,
    rois,
    temporal_radius=0.1,
    stc_rdm_metric="correlation",
    stc_rdm_params=dict(),
    rsa_metric="spearman",
    ignore_nan=False,
    y=None,
    n_folds=1,
    tmin=None,
    tmax=None,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA for a list of ROIs using MNE-Python source estimates.

    The output is a source estimate where the "signal" at each source point is
    the RSA, computed for a patch surrounding the source point. Source estimate
    objects can be either defined along a cortical surface (``SourceEstimate``
    objects) or volumetric (``VolSourceEstimate`` objects).  For surface source
    estimates, distances between vertices are measured in 2D space, namely as
    the length of the path along the surface from one vertex to another. For
    volume source estimates, distances are measured in 3D space as a straight
    line from one voxel to another.

    Parameters
    ----------
    stcs : list of mne.SourceEstimate | list of mne.VolSourceEstimate
        For each item, a source estimate for the brain activity.
    rdm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model RDM, see :func:`compute_rdm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model RDMs.

        Use :func:`compute_rdm` to compute RDMs.
    src : instance of mne.SourceSpaces
        The source space used by the source estimates specified in the `stcs`
        parameter.
    rois : list of mne.Label
        The spatial regions of interest (ROIs) to compute the RSA for. This
        needs to be specified as a list of ``mne.Label`` objects, such as
        returned by ``mne.read_annotations``.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    stc_rdm_metric : str
        The metric to use to compute the RDM for the source estimates. This can
        be any metric supported by the scipy.distance.pdist function. See also
        the ``stc_rdm_params`` parameter to specify and additional parameter
        for the distance function. Defaults to 'correlation'.
    stc_rdm_params : dict
        Extra arguments for the distance metric used to compute the RDMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing
        the distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Alternatively, you can pass a Scikit-Learn cross validator object (e.g.
        ``sklearn.model_selection.KFold``) to assert fine-grained control over
        how folds are created.
        Defaults to 1 (no cross-validation).
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    data : ndarray, shape (n_rois, n_times) | list of ndarray
        The correlation values for each ROI. When temporal_radius is set to
        None, there will be time dimension. When multiple models have been
        supplied, a list will be returned containing the RSA results for each
        model.
    stc : SourceEstimate | list of SourceEstimate
        The correlation values for each ROI, backfilled into a full
        SourceEstimate object. Each vertex belonging to the same ROI will have
        the same values. When temporal_radius is set to None, there will only
        be one time point. When multiple models have been supplied, a list will
        be returned containing the RSA results for each model.

    See Also
    --------
    compute_rdm
    """  # noqa E501
    # Check for compatibility of the source estimates and the model features
    one_model = type(rdm_model) is np.ndarray
    if one_model:
        rdm_model = [rdm_model]

    # Check for compatibility of the stcs and the model features
    for rdm in rdm_model:
        n_items = _n_items_from_rdm(rdm)
        if len(stcs) != n_items and y is None:
            raise ValueError(
                "The number of source estimates (%d) should be equal to the "
                "number of items in `rdm_model` (%d). Alternatively, use "
                "the `y` parameter to assign source estimates to items."
                % (len(stcs), n_items)
            )
        if y is not None and len(np.unique(y)) != n_items:
            raise ValueError(
                "The number of items in `rdm_model` (%d) does not match "
                "the number of items encoded in the `y` matrix (%d)."
                % (n_items, len(np.unique(y)))
            )

    src = _check_stcs_compatibility(stcs, src)

    if temporal_radius is not None:
        # Convert the temporal radius to samples
        temporal_radius = int(temporal_radius // stcs[0].tstep)

        if temporal_radius < 1:
            raise ValueError("Temporal radius is less than one sample.")

    samples_from, samples_to = _tmin_tmax_to_indices(stcs[0].times, tmin, tmax)

    # Convert the labels to data indices
    roi_inds = [vertex_selection_to_indices(stcs[0].vertices, roi) for roi in rois]

    # Perform the RSA
    X = np.array([stc.data for stc in stcs])
    patches = searchlight(
        X.shape,
        spatial_radius=roi_inds,
        temporal_radius=temporal_radius,
        samples_from=samples_from,
        samples_to=samples_to,
    )
    data = rsa_array(
        X,
        rdm_model,
        patches,
        data_rdm_metric=stc_rdm_metric,
        data_rdm_params=stc_rdm_params,
        rsa_metric=rsa_metric,
        ignore_nan=ignore_nan,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    # Pack the result in SourceEstimate objects
    subject = stcs[0].subject
    tmin = _construct_tmin(stcs[0].times, samples_from, samples_to, temporal_radius)
    tstep = stcs[0].tstep
    if one_model:
        stc = backfill_stc_from_rois(
            data, rois, src, tmin=tmin, tstep=tstep, subject=subject
        )
    else:
        stc = [
            backfill_stc_from_rois(
                data[..., i], rois, src, tmin=tmin, tstep=tstep, subject=subject
            )
            for i in range(data.shape[-1])
        ]

    return data, stc


@verbose
def rsa_nifti(
    image,
    rdm_model,
    spatial_radius=0.01,
    image_rdm_metric="correlation",
    image_rdm_params=dict(),
    rsa_metric="spearman",
    ignore_nan=False,
    y=None,
    n_folds=1,
    roi_mask=None,
    brain_mask=None,
    n_jobs=1,
    verbose=False,
):
    """Perform RSA in a searchlight pattern on Nibabel Nifti-like images.

    The output is a 3D Nifti image where the data at each voxel is is
    the RSA, computed for a patch surrounding the voxel.

    .. versionadded:: 0.4

    Parameters
    ----------
    image : 4D Nifti-like image
        The Nitfi image data. The 4th dimension must contain the images
        for each item.
    rdm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model RDM, see :func:`compute_rdm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model RDMs.

        Use :func:`compute_rdm` to compute RDMs.
    spatial_radius : float
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch.
        Defaults to 0.01.
    image_rdm_metric : str
        The metric to use to compute the RDM for the data. This can be
        any metric supported by the scipy.distance.pdist function. See also the
        ``image_rdm_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    image_rdm_params : dict
        Extra arguments for the distance metric used to compute the RDMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the RDMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    ignore_nan : bool
        Whether to treat NaN's as missing values and ignore them when computing
        the distance metric. Defaults to ``False``.

        .. versionadded:: 0.8
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Alternatively, you can pass a Scikit-Learn cross validator object (e.g.
        ``sklearn.model_selection.KFold``) to assert fine-grained control over
        how folds are created.
        Defaults to 1 (no cross-validation).
    roi_mask : 3D Nifti-like image | None
        When set, searchlight patches will only be generated for the subset of
        voxels with non-zero values in the given mask. This is useful for
        restricting the analysis to a region of interest (ROI). Note that while
        the center of the patches are all within the ROI, the patch itself may
        extend beyond the ROI boundaries.
        Defaults to ``None``, in which case patches for all voxels are
        generated.
    brain_mask : 3D Nifti-like image | None
        When set, searchlight patches are restricted to only contain voxels
        with non-zero values in the given mask. This is useful for make sure
        only information from inside the brain is used. In contrast to the
        `roi_mask`, searchlight patches will not use data outside of this mask.
        Defaults to ``None``, in which case all voxels are included in the
        analysis.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    rsa_results : 3D Nifti1Image | list of 3D Nifti1Image
        The correlation values for each searchlight patch. When multiple models
        have been supplied, a list will be returned containing the RSA results
        for each model.

    See Also
    --------
    compute_rdm

    """
    # Check for compatibility of the source estimates and the model features
    one_model = type(rdm_model) is np.ndarray
    if one_model:
        rdm_model = [rdm_model]

    if (
        not isinstance(image, tuple(nib.imageclasses.all_image_classes))
        or image.ndim != 4
    ):
        raise ValueError("The image data must be 4-dimensional Nifti-like " "images")

    # Check for compatibility of the BOLD images and the model features
    for rdm in rdm_model:
        n_items = _n_items_from_rdm(rdm)
        if image.shape[3] != n_items and y is None:
            raise ValueError(
                "The number of images (%d) should be equal to the "
                "number of items in `rdm_model` (%d). Alternatively, use "
                "the `y` parameter to assign evokeds to items."
                % (image.shape[3], n_items)
            )
        if y is not None and len(np.unique(y)) != n_items:
            raise ValueError(
                "The number of items in `rdm_model` (%d) does not match "
                "the number of items encoded in the `y` matrix (%d)."
                % (n_items, len(np.unique(y)))
            )

    # Get data as (n_items x n_voxels)
    X = image.get_fdata().reshape(-1, image.shape[3]).T

    # Find voxel positions
    voxels = np.array(list(np.ndindex(image.shape[:-1])))
    voxel_loc = voxels @ image.affine[:3, :3]
    voxel_loc /= 1000  # convert position from mm to meters

    # Apply masks
    result_mask = np.ones(image.shape[:3], dtype=bool)
    if brain_mask is not None:
        if brain_mask.ndim != 3 or brain_mask.shape != image.shape[:3]:
            raise ValueError(
                "Brain mask must be a 3-dimensional Nifi-like "
                "image with the same dimensions as the data "
                "image"
            )
        brain_mask = brain_mask.get_fdata() != 0
        result_mask &= brain_mask
        brain_mask = brain_mask.ravel()
        X = X[:, brain_mask]
        voxel_loc = voxel_loc[brain_mask]
    if roi_mask is not None:
        if roi_mask.ndim != 3 or roi_mask.shape != image.shape[:3]:
            raise ValueError(
                "ROI mask must be a 3-dimensional Nifi-like "
                "image with the same dimensions as the data "
                "image"
            )
        roi_mask = roi_mask.get_fdata() != 0
        result_mask &= roi_mask
        roi_mask = roi_mask.ravel()
        if brain_mask is not None:
            roi_mask = roi_mask[brain_mask]
        roi_mask = np.flatnonzero(roi_mask)

    # Compute distances between voxels
    logger.info("Computing distances...")
    from sklearn.neighbors import NearestNeighbors

    nn = NearestNeighbors(radius=spatial_radius, n_jobs=n_jobs).fit(voxel_loc)
    dist = nn.radius_neighbors_graph(mode="distance")

    # Perform the RSA
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=None,
        sel_series=roi_mask,
    )
    rsa_result = rsa_array(
        X,
        rdm_model,
        patches,
        data_rdm_metric=image_rdm_metric,
        data_rdm_params=image_rdm_params,
        rsa_metric=rsa_metric,
        ignore_nan=ignore_nan,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    if one_model:
        data = np.zeros(image.shape[:3])
        data[result_mask] = rsa_result
        return nib.Nifti1Image(data, image.affine, image.header)
    else:
        results = []
        for i in range(rsa_result.shape[-1]):
            data = np.zeros(image.shape[:3])
            data[result_mask] = rsa_result[:, i]
            results.append(nib.Nifti1Image(data, image.affine, image.header))
        return results


@verbose
def rdm_nifti(
    image,
    spatial_radius=0.01,
    dist_metric="correlation",
    dist_params=dict(),
    y=None,
    n_folds=1,
    roi_mask=None,
    brain_mask=None,
    n_jobs=1,
    verbose=False,
):
    """Generate RDMs in a searchlight pattern on Nibabel Nifty-like images.

    RDMs are computed using a patch surrounding each voxel.

    .. versionadded:: 0.4

    Parameters
    ----------
    image : 4D Nifti-like image
        The Nitfi image data. The 4th dimension must contain the images
        for each item.
    spatial_radius : float
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch.
        Defaults to 0.01.
    dist_metric : str
        The metric to use to compute the RDM for the data. This can be
        any metric supported by the scipy.distance.pdist function. See also the
        ``dist_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the RDMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | sklearn.model_selection.BaseCrollValidator | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Alternatively, you can pass a Scikit-Learn cross validator object (e.g.
        ``sklearn.model_selection.KFold``) to assert fine-grained control over
        how folds are created.
        Defaults to 1 (no cross-validation).
    roi_mask : 3D Nifti-like image | None
        When set, searchlight patches will only be generated for the subset of
        voxels with non-zero values in the given mask. This is useful for
        restricting the analysis to a region of interest (ROI). Note that while
        the center of the patches are all within the ROI, the patch itself may
        extend beyond the ROI boundaries.
        Defaults to ``None``, in which case patches for all voxels are
        generated.
    brain_mask : 3D Nifti-like image | None
        When set, searchlight patches are restricted to only contain voxels
        with non-zero values in the given mask. This is useful for make sure
        only information from inside the brain is used. In contrast to the
        `roi_mask`, searchlight patches will not use data outside of this mask.
        Defaults to ``None``, in which case all voxels are included in the
        analysis.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Yields
    ------
    rdm : ndarray, shape (n_items, n_items)
        A RDM for each searchlight patch.

    """
    if (
        not isinstance(image, tuple(nib.imageclasses.all_image_classes))
        or image.ndim != 4
    ):
        raise ValueError("The image data must be 4-dimensional Nifti-like " "images")

    # Get data as (n_items x n_voxels)
    X = image.get_fdata().reshape(-1, image.shape[3]).T

    # Find voxel positions
    voxels = np.array(list(np.ndindex(image.shape[:-1])))
    voxel_loc = voxels @ image.affine[:3, :3]
    voxel_loc /= 1000  # convert position from mm to meters

    # Apply masks
    result_mask = np.ones(image.shape[:3], dtype=bool)
    if brain_mask is not None:
        if brain_mask.ndim != 3 or brain_mask.shape != image.shape[:3]:
            raise ValueError(
                "Brain mask must be a 3-dimensional Nifi-like "
                "image with the same dimensions as the data "
                "image"
            )
        brain_mask = brain_mask.get_fdata() != 0
        result_mask &= brain_mask
        brain_mask = brain_mask.ravel()
        X = X[:, brain_mask]
        voxel_loc = voxel_loc[brain_mask]
    if roi_mask is not None:
        if roi_mask.ndim != 3 or roi_mask.shape != image.shape[:3]:
            raise ValueError(
                "ROI mask must be a 3-dimensional Nifi-like "
                "image with the same dimensions as the data "
                "image"
            )
        roi_mask = roi_mask.get_fdata() != 0
        result_mask &= roi_mask
        roi_mask = roi_mask.ravel()
        if brain_mask is not None:
            roi_mask = roi_mask[brain_mask]
        roi_mask = np.flatnonzero(roi_mask)

    # Compute distances between voxels
    logger.info("Computing distances...")
    from sklearn.neighbors import NearestNeighbors

    nn = NearestNeighbors(radius=spatial_radius, n_jobs=n_jobs).fit(voxel_loc)
    dist = nn.radius_neighbors_graph(mode="distance")

    # Compute RDMs
    patches = searchlight(
        X.shape,
        dist=dist,
        spatial_radius=spatial_radius,
        temporal_radius=None,
        sel_series=roi_mask,
    )
    yield from rdm_array(
        X,
        patches,
        dist_metric=dist_metric,
        dist_params=dist_params,
        y=y,
        n_folds=n_folds,
        n_jobs=n_jobs,
    )


def _check_stcs_compatibility(stcs, src):
    """Check for compatibility of the source estimates and source space."""
    if src.kind == "volume" and not isinstance(stcs[0], mne.VolSourceEstimate):
        raise ValueError(
            "Volume source estimates provided, but not a volume source space "
            f"(src.kind={src.kind})."
        )
    if src.kind == "volume":
        if np.any(stcs[0].vertices != src[0]["vertno"]):
            src = _restrict_src_to_vertices(src, stcs[0].vertices)
        for stc in stcs:
            if np.any(stc.vertices != src[0]["vertno"]):
                raise ValueError("Not all source estimates have the same vertices.")
    else:
        for src_hemi, stc_hemi_vertno in zip(src, stcs[0].vertices):
            if np.any(stc_hemi_vertno != src_hemi["vertno"]):
                src = _restrict_src_to_vertices(src, stcs[0].vertices)
        for stc in stcs:
            for src_hemi, stc_hemi_vertno in zip(src, stcs[0].vertices):
                if np.any(stc_hemi_vertno != src_hemi["vertno"]):
                    raise ValueError("Not all source estimates have the same vertices.")

    times = stcs[0].times
    for stc in stcs:
        if np.any(stc.times != times):
            raise ValueError("Not all source estimates have the same time points.")
    return src


def _get_distance_matrix(src, dist_lim, n_jobs=1):
    """Get vertex-to-vertex distance matrix from source space.

    During inverse computation, the source space was downsampled (i.e. using
    ico4). Construct vertex-to-vertex distance matrices using only the
    vertices that are defined in the source solution.

    Parameters
    ----------
    src : mne.SourceSpaces
        The source space to get the distance matrix for.
    dist_lim : float
        Maximum distance required. We don't care about distances beyond this
        maximum.
    n_jobs : int
        Number of CPU cores to use if distance computation is necessary.
        Defaults to 1.

    Returns
    -------
    dist : ndarray (n_vertices, n_vertices)
        The vertex-to-vertex distance matrix.

    """
    dist = []

    # Check if distances have been pre-computed in the given source space. Give
    # a warning if the pre-computed distances may have had a too limited
    # dist_lim setting.
    needs_distance_computation = False
    for hemi in src:
        if "dist" not in hemi or hemi["dist"] is None:
            needs_distance_computation = True
        else:
            if hemi["dist_limit"] != np.inf and hemi["dist_limit"][0] < dist_lim:
                warn(
                    f"Source space has pre-computed distances, but all "
                    f"distances are smaller than the searchlight radius "
                    f"({dist_lim}). You may want to consider recomputing "
                    f"the source space distances using the "
                    f"mne.add_source_space_distances function."
                )

    if needs_distance_computation:
        if dist_lim is None:
            dist_lim = np.inf
        if src.kind == "volume":
            src = _add_volume_source_space_distances(src, dist_lim)
        else:
            src = mne.add_source_space_distances(src, dist_lim, n_jobs=n_jobs)

    for hemi in src:
        inuse = np.flatnonzero(hemi["inuse"])
        dist.append(hemi["dist"][np.ix_(inuse, inuse)].toarray())

    # Collect the distances in a single matrix
    dist = block_diag(*dist)
    dist[dist == 0] = np.inf  # Across hemisphere distance is infinity
    dist.flat[:: dist.shape[0] + 1] = 0  # Distance to yourself is zero

    return dist


def _add_volume_source_space_distances(src, dist_limit):
    """Compute the distance between voxels in a volume source space.

    Operates in-place!

    Code is mostly taken from `mne.add_source_space_distances`.

    Parameters
    ----------
    src : instance of mne.SourceSpaces
        The volume source space to compute the voxel-wise distances for.
    dist_limit : float
        The maximum distance (in meters) to consider. Voxels that are further
        apart than this distance will have a distance of infinity. Use this to
        reduce computation time.

    Returns
    -------
    src : instance of mne.SourceSpaces
        The volume source space, now with the 'dist' and 'dist_limit' fields
        set.

    """
    # Lazy import to not have to load the huge scipy module every time mne_rsa
    # get's loaded.
    from scipy.sparse import csr_matrix

    assert src.kind == "volume"
    n_sources = src[0]["np"]
    neighbors = np.array(src[0]["neighbor_vert"])
    rows, cols = np.nonzero(neighbors != -1)
    cols = neighbors[(rows, cols)]
    dist = np.linalg.norm(src[0]["rr"][rows, :] - src[0]["rr"][cols, :], axis=1)
    con_matrix = csr_matrix((dist, (rows, cols)), shape=(n_sources, n_sources))
    dist = mne.source_space._source_space._do_src_distances(
        con_matrix, src[0]["vertno"], np.arange(src[0]["nuse"]), dist_limit
    )[0]
    d = dist.ravel()  # already float32
    idx = d > 0
    d = d[idx]
    i, j = np.meshgrid(src[0]["vertno"], src[0]["vertno"])
    i = i.ravel()[idx]
    j = j.ravel()[idx]
    src[0]["dist"] = csr_matrix((d, (i, j)), shape=(n_sources, n_sources))
    src[0]["dist_limit"] = np.array([dist_limit], "float32")
    return src


def backfill_stc_from_rois(values, rois, src, tmin=0, tstep=1, subject=None):
    """Backfill the ROI values into a full mne.SourceEstimate object.

    Each vertex belonging to the same region of interest (ROI) will have the
    sample value.

    Parameters
    ----------
    values : ndarray, shape (n_rois, ...)
        For each ROI, either a single value or a timecourse of values.
    rois : list of mne.Label
        The spatial regions of interest (ROIs) to compute the RSA for. This
        needs to be specified as a list of ``mne.Label`` objects, such as
        returned by ``mne.read_annotations``.
    src : instance of mne.SourceSpaces
        The source space used by the source estimates specified in the `stcs`
        parameter.
    tmin : float
        Time corrsponding to the first sample.
    tstep : float
        Difference in time between two samples.
    subject : str | None
        The name of the FreeSurfer subject.

    Returns
    -------
    stc : mne.SourceEstimate
        The backfilled source estimate object.

    """
    values = np.asarray(values)
    if values.ndim == 1:
        n_samples = 1
    else:
        n_samples = values.shape[1]
    data = np.zeros((src[0]["nuse"] + src[1]["nuse"], n_samples))
    verts_lh = src[0]["vertno"]
    verts_rh = src[1]["vertno"]
    for roi, rsa_timecourse in zip(rois, values):
        roi = roi.copy().restrict(src)
        if roi.hemi == "lh":
            roi_ind = np.searchsorted(verts_lh, roi.vertices)
        else:
            roi_ind = np.searchsorted(verts_rh, roi.vertices)
            roi_ind += src[0]["nuse"]
        for ind in roi_ind:
            data[ind] = rsa_timecourse
    return mne.SourceEstimate(
        data, vertices=[verts_lh, verts_rh], tmin=tmin, tstep=tstep, subject=subject
    )


def vertex_selection_to_indices(vertno, sel_vertices):
    """Unify across different ways of selecting vertices."""
    if isinstance(sel_vertices, mne.Label):
        sel_vertices = [sel_vertices]
    if not isinstance(sel_vertices, list):
        raise ValueError(
            "Invalid type for sel_vertices. It should be an mne.Label, a list of "
            f"mne.Label's or a list of numpy arrays, but {type(sel_vertices)} was "
            "provided."
        )
    if len(sel_vertices) == 0:
        raise ValueError("Empty list provided for sel_vertices.")

    # Deal with mne.Label objects
    if isinstance(sel_vertices[0], mne.Label):
        labels = sel_vertices
        sel_vertices = [[] for _ in range(len(vertno))]
        for label in labels:
            if label.hemi == "lh":
                sel_vertices[0].extend(label.get_vertices_used(vertno[0]).tolist())
            else:
                if len(vertno) == 1:
                    raise ValueError(
                        "one of the Label's provided for sel_vertices defines vertices "
                        "on the right hemisphere. However, the provided "
                        "SourceEstimate's only have vertices in the left hemisphere."
                    )
                sel_vertices[1].extend(label.get_vertices_used(vertno[1]).tolist())

    # At this point, sel_vertices should be an array of vertex numbers.
    # Convert them to indices for the .data array.
    if len(sel_vertices) != len(vertno):
        raise ValueError(
            f"sel_vertices should be a list with {len(vertno)} elements: one for "
            "each hemisphere (volume source spaces have one hemisphere.)"
        )
    sel_series = []
    for hemi, (hemi_vertno, hemi_sel_vertno) in enumerate(zip(vertno, sel_vertices)):
        if len(hemi_sel_vertno) == 0:
            continue
        sel_inds = np.searchsorted(hemi_vertno, hemi_sel_vertno)
        if not np.array_equal(hemi_vertno[sel_inds], hemi_sel_vertno):
            raise ValueError("Some selected vertices are not present in the data.")
        if hemi > 0:
            sel_inds += len(vertno[hemi - 1])
        sel_series.append(sel_inds)
    return np.unique(np.hstack(sel_series))


def vertex_indices_to_numbers(vertno, vert_ind):
    """Convert vertex indices to vertex numbers."""
    min_vert_ind = 0
    sel_vert_no = [[] for _ in vertno]
    for hemi, hemi_vertno in enumerate(vertno):
        if hemi > 0:
            min_vert_ind += len(vertno[hemi - 1])
        max_vert_ind = min_vert_ind + len(hemi_vertno)
        hemi_vert_ind = vert_ind[(vert_ind >= min_vert_ind) & (vert_ind < max_vert_ind)]
        hemi_vert_ind -= min_vert_ind
        sel_vert_no[hemi] = hemi_vertno[hemi_vert_ind]
    return sel_vert_no


@verbose
def _restrict_src_to_vertices(src, vertno, verbose=None):
    """Restrict a source space to the given vertices.

    Parameters
    ----------
    src: instance of SourceSpaces
        The source space to be restricted.
    vertno : tuple of lists (vertno_lh, vertno_rh)
        For each hemisphere, the vertex numbers to keep.
    verbose : bool | str | int | None
        If not None, override default verbose level (see :func:`mne.verbose`
        and :ref:`Logging documentation <tut_logging>` for more).

    Returns
    -------
    src_out : instance of SourceSpaces
        The restricted source space.

    """
    src_out = deepcopy(src)

    vert_no_lh, vert_no_rh = vertno
    if not (
        np.all(np.in1d(vert_no_lh, src[0]["vertno"]))
        and np.all(np.in1d(vert_no_rh, src[1]["vertno"]))
    ):
        raise ValueError("One or more vertices were not present in the source space.")
    logger.info(
        "Restricting source space to {n_keep} out of {n_total} vertices.".format(
            n_keep=len(vert_no_lh) + len(vert_no_rh),
            n_total=src[0]["nuse"] + src[1]["nuse"],
        )
    )

    for hemi, verts in zip(src_out, (vert_no_lh, vert_no_rh)):
        # Ensure vertices are in sequential order
        verts = np.sort(verts)

        # Restrict the source space
        hemi["vertno"] = verts
        hemi["nuse"] = len(verts)
        hemi["inuse"] = hemi["inuse"].copy()
        hemi["inuse"].fill(0)
        if hemi["nuse"] > 0:  # Don't use empty array as index
            hemi["inuse"][verts] = 1
        hemi["use_tris"] = np.array([[]], int)
        hemi["nuse_tri"] = np.array([0])

    return src_out



"""Unit tests for the creation of RDMs."""

import numpy as np
import pytest
from mne_rsa import compute_rdm, compute_rdm_cv, pick_rdm, rdm_array, searchlight
from mne_rsa.rdm import _ensure_condensed, _n_items_from_rdm
from numpy.testing import assert_allclose, assert_equal


class TestDsm:
    """Test computing a RDM."""

    def test_basic(self):
        """Test basic invocation of compute_rdm."""
        data = np.array([[1, 2, 3, 4], [1, 2, 3, 4]])
        rdm = compute_rdm(data)
        assert rdm.shape == (1,)
        assert_allclose(rdm, 0, atol=1e-15)

    def test_invalid_input(self):
        """Test giving invalid input to compute_rdm."""
        data = np.array([[1], [1]])
        with pytest.raises(ValueError, match="single feature"):
            compute_rdm(data, metric="correlation")

    def test_set_metric(self):
        """Test setting distance metric for computing RDMs."""
        data = np.array([[1, 2, 3, 4], [2, 4, 6, 8]])
        rdm = compute_rdm(data, metric="euclidean")
        assert rdm.shape == (1,)
        assert_allclose(rdm, 5.477226)


class TestDsmCV:
    """Test computing a RDM with cross-validation."""

    def test_basic(self):
        """Test basic invocation of compute_rdm_cv."""
        data = np.array([[[1, 2, 3, 4], [1, 2, 3, 4]], [[1, 2, 3, 4], [1, 2, 3, 4]]])
        rdm = compute_rdm_cv(data)
        assert rdm.shape == (1,)
        assert_allclose(rdm, 0, atol=1e-15)

    def test_invalid_input(self):
        """Test giving invalid input to compute_rdm."""
        data = np.array([[[1], [1]]])
        with pytest.raises(ValueError, match="single feature"):
            compute_rdm_cv(data, metric="correlation")

    def test_set_metric(self):
        """Test setting distance metric for computing RDMs."""
        data = np.array([[[1, 2, 3, 4], [2, 4, 6, 8]], [[1, 2, 3, 4], [2, 4, 6, 8]]])
        rdm = compute_rdm_cv(data, metric="euclidean")
        assert rdm.shape == (1,)
        assert_allclose(rdm, 5.477226)


class TestEnsureCondensed:
    """Test the _ensure_condensed function."""

    def test_basic(self):
        """Test basic invocation of _ensure_condensed."""
        rdm = _ensure_condensed(
            np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]]), var_name="test"
        )
        assert rdm.shape == (3,)
        assert_equal(rdm, [1, 2, 3])

    def test_list(self):
        """Test invocation of _ensure_condensed on a list."""
        full = [
            np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]]),
            np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]]),
        ]
        rdm = _ensure_condensed(full, var_name="full")
        assert len(rdm) == 2
        assert rdm[0].shape == (3,)
        assert rdm[1].shape == (3,)
        assert_equal(rdm, [[1, 2, 3], [1, 2, 3]])

    def test_condensed(self):
        """Test invocation of _ensure_condensed on already condensed RDM."""
        rdm = _ensure_condensed(np.array([1, 2, 3]), var_name="test")
        assert rdm.shape == (3,)
        assert_equal(rdm, [1, 2, 3])

    def test_invalid(self):
        """Test _ensure_condensed with invalid inputs."""
        # Not a square matrix
        with pytest.raises(ValueError, match="square matrix"):
            _ensure_condensed(np.array([[0, 1], [1, 0], [2, 3]]), var_name="test")

        # Too many dimensions
        with pytest.raises(ValueError, match="Invalid dimensions"):
            _ensure_condensed(np.array([[[[[0, 1, 2, 3]]]]]), var_name="test")

        # Invalid type
        with pytest.raises(TypeError, match="NumPy array"):
            _ensure_condensed([1, 2, 3], var_name="test")


class TestNItemsFromRDM:
    """Test the _n_items_from_rdm function."""

    def test_basic(self):
        """Test basic invocation of _n_items_from_rdm."""
        assert _n_items_from_rdm(np.array([1, 2, 3])) == 3
        assert _n_items_from_rdm(np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])) == 3
        with pytest.raises(ValueError):
            _n_items_from_rdm(np.array([[[1, 2, 3]]]))


class TestSelectItemsFromRDM:
    """Test the _select_items_from_rdm function."""

    def test_basic(self):
        """Test basic invocation of _select_items_from_rdm."""
        assert_equal(pick_rdm(np.array([1, 2, 3]), 1), np.array([]))
        assert_equal(
            pick_rdm(np.array([1, 2, 3]), slice(None)),
            np.array([1, 2, 3]),
        )
        assert_equal(pick_rdm(np.array([1, 2, 3]), [0, 1]), np.array([1]))
        assert_equal(pick_rdm(np.array([1, 2, 3]), [0, 2]), np.array([2]))
        assert_equal(pick_rdm(np.array([1, 2, 3]), [1, 2]), np.array([3]))
        assert_equal(
            pick_rdm(np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]]), [1, 2]),
            np.array([[0, 3], [3, 0]]),
        )
        with pytest.raises(ValueError):
            pick_rdm(np.array([[[1, 2, 3]]]), 1)


class TestDsmsSearchlight:
    """Test computing RDMs with searchlight patches."""

    def test_temporal(self):
        """Test computing RDMs using a temporal searchlight."""
        data = np.array([[1, 2, 3, 4], [1, 2, 3, 4]])
        patches = searchlight(data.shape, temporal_radius=1)
        rdms = rdm_array(data, patches, dist_metric="euclidean")
        assert len(rdms) == len(patches)
        assert rdms.shape == (2, 1)
        assert_equal(list(rdms), [0, 0])

    def test_spatial(self):
        """Test computing RDMs using a spatial searchlight."""
        dist = np.array([[0, 1, 2, 3], [1, 0, 1, 2], [2, 1, 0, 1], [3, 2, 1, 0]])
        data = np.array([[1, 2, 3, 4], [1, 2, 3, 4]])
        patches = searchlight(data.shape, dist, spatial_radius=1)
        rdms = rdm_array(data, patches, dist_metric="euclidean")
        assert len(rdms) == len(patches)
        assert rdms.shape == (4, 1)
        assert_equal(list(rdms), [0, 0, 0, 0])

    def test_spatio_temporal(self):
        """Test computing RDMs using a spatio-temporal searchlight."""
        data = np.array(
            [[[1, 2, 3], [2, 3, 4]], [[2, 3, 4], [3, 4, 5]], [[3, 4, 5], [4, 5, 6]]]
        )
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])
        patches = searchlight(data.shape, dist, spatial_radius=1, temporal_radius=1)
        rdms = rdm_array(data, patches, dist_metric="correlation")
        assert len(rdms) == len(patches)
        assert rdms.shape == (2, 1, 3)
        assert_allclose(list(rdms), [[0, 0, 0], [0, 0, 0]], atol=1e-15)

    def test_single_patch(self):
        """Test computing RDMs using a single searchlight patch."""
        data = np.array(
            [[[1, 2, 3], [2, 3, 4]], [[2, 3, 4], [3, 4, 5]], [[3, 4, 5], [4, 5, 6]]]
        )
        rdms = rdm_array(data, dist_metric="correlation")
        assert len(rdms) == 1
        assert rdms.shape == (3,)
        assert_allclose(list(rdms), [[0, 0, 0]], atol=1e-15)

    def test_crossvalidation(self):
        """Test computing RDMs using a searchlight and cross-validation."""
        data = np.array(
            [
                [[1, 2, 3], [2, 3, 4]],
                [[2, 3, 4], [3, 4, 5]],
                [[3, 4, 5], [4, 5, 6]],
                [[1, 2, 3], [2, 3, 4]],
                [[2, 3, 4], [3, 4, 5]],
                [[3, 4, 5], [4, 5, 6]],
            ]
        )
        dist = np.array([[0, 1, 2], [1, 0, 1], [2, 1, 0]])
        patches = searchlight(data.shape, dist, spatial_radius=1, temporal_radius=1)
        rdms = rdm_array(data, patches, y=[1, 2, 3, 1, 2, 3], n_folds=2)
        assert len(rdms) == len(patches)
        assert rdms.shape == (2, 1, 3)
        assert_allclose(list(rdms), [[0, 0, 0], [0, 0, 0]], atol=1e-15)

    def test_generator(self):
        """Test generator behavior when computing RDMs."""
        data = np.array([[1, 2, 3, 4], [1, 2, 3, 4]])
        patches = searchlight(data.shape, temporal_radius=1)
        rdms = rdm_array(data, patches, dist_metric="euclidean")
        assert len(rdms) == len(patches)
        assert next(rdms).shape == (1,)
        # test restart behavior
        rdm_list1 = [rdm for rdm in iter(rdms)]
        rdm_list2 = [rdm for rdm in iter(rdms)]
        assert len(rdm_list1) == len(rdm_list2) == 2



"""Unit tests for the creation of cross-validation folds."""

import numpy as np
import pytest
from mne_rsa.folds import _compute_item_means, _convert_to_one_hot, create_folds
from numpy.testing import assert_equal


class TestCreateFolds:
    """Test the create_folds function."""

    def test_basic(self):
        """Test basic invocation of create_folds."""
        data = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3])
        y = [1, 2, 3, 1, 2, 3, 1, 2, 3]

        folds = create_folds(data, y, n_folds=1)
        assert_equal(folds, [[2, 2, 2]])

        folds = create_folds(data, y, n_folds=2)
        assert_equal(folds, [[1.5, 1, 1.5], [3, 2.5, 3]])

        folds = create_folds(data, y, n_folds=3)
        assert_equal(folds, [[1, 1, 1], [2, 2, 2], [3, 3, 3]])

        folds = create_folds(data[:, np.newaxis], y, n_folds=3)
        assert_equal(folds, [[[1], [1], [1]], [[2], [2], [2]], [[3], [3], [3]]])

        # Default value for n_folds is maximum number of folds
        assert_equal(create_folds(data, y), create_folds(data, y, n_folds=3))

        # No folding when y=None
        folds = create_folds(data, y=None)
        assert_equal(folds, [data])

    def test_invalid_input(self):
        """Test passing invalid input to create_folds."""
        # Invalid y
        data = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3])
        y = [1, 2, 3]
        with pytest.raises(ValueError, match="length of y"):
            create_folds(data, y)


class TestConvertToOneHot:
    """Test the _convert_to_one_hot function."""

    def test_basic(self):
        """Test basic invocation of _convert_to_one_hot."""
        y = np.array([1, 2, 3, 1, 2, 3])
        y_one_hot = np.array(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 0], [0, 1, 0], [0, 0, 1]]
        )

        assert_equal(_convert_to_one_hot(y), y_one_hot)
        assert_equal(_convert_to_one_hot(y.tolist()), y_one_hot)
        assert_equal(_convert_to_one_hot(y[:, np.newaxis]), y_one_hot)
        assert_equal(_convert_to_one_hot(y_one_hot), y_one_hot)

    def test_invalid_input(self):
        """Test passing invalid input to _convert_to_one_hot."""
        with pytest.raises(ValueError, match="Wrong number of dimensions"):
            _convert_to_one_hot([[[1], [2], [3]]])


class TestComputeItemMeans:
    """Test the _compute_item_means function."""

    def test_basic(self):
        """Test basic invocation of _compute_item_means."""
        data = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3])
        y_one_hot = _convert_to_one_hot([1, 2, 3, 1, 2, 3, 1, 2, 3])
        assert_equal(_compute_item_means(data, y_one_hot), [2, 2, 2])
